# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
from datetime import datetime
import hashlib
from retention import Retention
import time
import os


FILE_PATH = os.path.abspath(os.path.dirname(__file__))


class MapReduceRetention():
    def __init__(self, slice_count):
        if not os.path.exists('{}/temp'.format(FILE_PATH)):
            os.mkdir('{}/temp'.format(FILE_PATH))
        self.slice_count = slice_count

    def _hash_id(self, id):
        id = id.encode(encoding='UTF-8', errors='strict')
        int_id = int(hashlib.sha1(id).hexdigest(), 16) % (10 ** 18)
        return int_id

    def _clear_dir(self, dirpath=None):
        if not dirpath:
            dirpath = '{}/temp'.format(FILE_PATH)
        if os.path.exists(dirpath):
            files = os.listdir(dirpath)
            for file in files:
                filepath = os.path.join(dirpath, file).replace("\\", '/')
                if os.path.isdir(filepath):
                    self._clear_dir(filepath)
                else:
                    os.remove(filepath)
            os.rmdir(dirpath)

    def map_origin_df_withhash(self, data):
        int_id = self._hash_id(data['user_id'])
        data['group'] = int_id % self.slice_count
        return data

    def map_origin_df_withouthash(self, data):
        data['group'] = data['user_id'] % self.slice_count
        return data

    #id_type num: 数字类型id str: 字符串类型id
    def slice_to_csv(self, source, mode = 'csv', id_type='num'):
        self.df_array = {}
        for i in range(0, self.slice_count):
            self.df_array[i] = pd.DataFrame()
        chunks = pd.read_csv(source, chunksize=100000)
        count = 0
        for df in chunks:
            df['group'] = -1
            if id_type == 'str':
                df = df.apply(self.map_origin_df_withhash, axis=1)
            elif id_type == 'num':
                df = df.apply(self.map_origin_df_withouthash, axis=1)
            for group in range(0, self.slice_count):
                self.df_array[group] = df[df['group'] == group]
                if count == 0:
                    self.df_array[group].to_csv(
                        '{}/temp/map-{}.csv'.format(FILE_PATH, group), index=False, encoding='utf_8_sig')
                else:
                    self.df_array[group].to_csv(
                        '{}/temp/map-{}.csv'.format(FILE_PATH, group), mode='a', header=False, index=False, encoding='utf_8_sig')
            count += 1

    # map_task中的rate只能是sum或者count
    def map_task(self, period='month', loop=None, format='%Y-%m-%d %H:%M:%S', mode='count', sum_word=None):
        for group in range(0, self.slice_count):
            Retention(file_name='{}/temp/map-{}.csv'.format(FILE_PATH, group), period=period, loop=loop, format=format, mode=mode, sum_word=sum_word, write_name='{}/temp/map-retention{}.csv'.format(FILE_PATH, group))

    #计算rate, shape: tri三角形, 对应三角形矩阵, rect, 对应矩形矩阵
    def cal_rate(self, matrix, shape='tri'):
        origin_df = matrix.copy(deep=True)
        if shape == 'tri':
            for i in range(0, matrix.shape[0]):
                ref = origin_df.iloc[[i], [i]]
                for j in range(i, matrix.shape[1]):
                    matrix.iloc[[i], [j]] = matrix.iloc[[i], [j]].values[0][0] / ref.values[0][0]
        elif shape == 'rect':
            for i in range(0, matrix.shape[0]):
                ref = origin_df.iloc[[i], [0]]
                for j in range(0, matrix.shape[1]):
                    matrix.iloc[[i], [j]] = matrix.iloc[[i], [j]].values[0][0] / ref.values[0][0]
        return matrix

    def cal_rate_tri(self, matrix):
        origin_df = matrix.copy(deep=True)
        for i in range(0, matrix.shape[0]):
            ref = origin_df.iloc[[i], [i]]
            for j in range(i, matrix.shape[1]):
                matrix.iloc[[i], [j]] = matrix.iloc[[i], [j]].values[0][0] / ref.values[0][0]
        return matrix

    def reduce_task(self, mode='count', shape='tri', write_name='rention.csv'):
        sum_df = pd.DataFrame()
        for group in range(0, self.slice_count):
            df = pd.read_csv(
                '{}/temp/map-retention{}.csv'.format(FILE_PATH, group), index_col=0)
            for index, row in df.iterrows():
                for column in row.index:
                    if not column in sum_df.columns:
                        sum_df[column] = None
                    if not index in sum_df.index:
                        sum_df.loc[index] = np.zeros(len(sum_df.columns))
                    '''赋值'''
                    if not df[column].loc[index]:
                        sum_df[column].loc[index] = None
                    elif not sum_df[column].loc[index]:
                        sum_df[column].loc[index] = df[column].loc[index]
                    else:
                        sum_df[column].loc[index] += df[column].loc[index]
        sum_df.sort_index(inplace=True)
        sum_df = pd.DataFrame(
            sum_df.values.T, index=sum_df.columns, columns=sum_df.index)
        sum_df.sort_index(inplace=True)
        sum_df = pd.DataFrame(
            sum_df.values.T, index=sum_df.columns, columns=sum_df.index)
        if mode == 'rate':
            sum_df = self.cal_rate(sum_df, shape)
        sum_df.to_csv(write_name, encoding='utf_8_sig')
        # print(sum_df)


if __name__ == '__main__':
    ''' 切成10个小块'''
    mrr = MapReduceRetention(10)
    ''' 将任务分发到不同的csv中'''
    # mrr.slice_to_csv('data.csv')


    # eg1: 矩形月人数
    # ''' 拆分计算任务'''
    # mrr.map_task(loop=[0, 1, 2], period='month', mode='count')
    # '''合并计算任务'''
    # mrr.reduce_task(mode='rate', shape='react', write_name='eg-loop-month-rate.csv')

    # eg1: 矩形月销售额累加
    # ''' 拆分计算任务'''
    # mrr.map_task(loop=[0, 1, 2], period='month', mode='sum', sum_word='order_amount')
    # '''合并计算任务'''
    # mrr.reduce_task(mode='count', shape='react', write_name='eg-loop-month-rate.csv')

    '''清理目录'''
    mrr._clear_dir()
