# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
import logging
from datetime import datetime
'''
计算留存率/留存数/留存求和
参数: {
    period: n // 日、周、月、年，或指定时间间隔（一个时间段） day, month, year, week, quarter, halfyear
    loop: [] || None // 周期的循环次数，不指定则循环到当前时间
    mode: user--用户留存次数/留存率 action--购买行为留存数/留存率 sum--复购留存用户对某指标求和/比例,要同时指定求和列名sum_word
}
数据源格式:
user_id active_time
'''


class Retention():
    def __init__(self, file_name='data.csv', active_time='active_time', user_id='user_id', period='day', loop=None, format='%Y-%m-%d', mode='rate', sum_word=None, write_name=None):
        self.file_name = file_name
        self.active_time = active_time
        self.user_id = user_id
        self.write_name = write_name
        self.mode = mode
        self.sum_word = sum_word
        self.period = period
        self.loop = loop
        self.format = format
        self.data = pd.read_csv(self.file_name, encoding='utf_8_sig')
        self.data[self.active_time] = self.data[self.active_time].map(self._map_time)
        self._new_users()
    # 给每一条数据都新增一列时间属性，为周期时间

    def _map_time(self, data):
        time = datetime.strptime(data, self.format)
        quarter_map = {
            1: 1,
            2: 1,
            3: 1,
            4: 2,
            5: 2,
            6: 2,
            7: 3,
            8: 3,
            9: 3,
            10: 4,
            11: 4,
            12: 4
        }
        halfyear_map = {
            1: 1,
            2: 1,
            3: 1,
            4: 1,
            5: 1,
            6: 1,
            7: 2,
            8: 2,
            9: 2,
            10: 2,
            11: 2,
            12: 2
        }
        quarter = '%02d' % quarter_map[time.month]
        halfyear = '%02d' % halfyear_map[time.month]
        day = '%02d' % time.day
        week = '%02d' % datetime(time.year,time.month,time.day).isocalendar()[1]
        month = '%02d' % time.month
        year = '%04d' % time.year
        if self.period == 'day':
            return '{}-{}-{}'.format(year, month, day)
        elif self.period == 'week':
            return '{}-{}'.format(year, week)
        elif self.period == 'month':
            return '{}-{}'.format(year, month)
        elif self.period == 'quarter':
            return '{}-{}'.format(year, quarter)
        elif self.period == 'halfyear':
            return '{}-{}'.format(year, halfyear)
        elif self.period == 'year':
            return '{}'.format(year)
    # 找到每一组的新增用户，排序

    def _new_users(self):
        self.init_users = self.data.groupby(
            [self.user_id])[self.active_time].min().reset_index(self.user_id)
        self.retention()

    def _remained(self, df_origin=None, df_current=None, first_action_count=None, first_action_sum=None):
        if df_origin.shape[0]:
            intersection = pd.merge(df_origin, df_current, on=self.user_id)
            if self.mode == 'user':
                remained = len(intersection.groupby([self.user_id]))/len(df_origin.groupby([self.user_id]))
                return [len(intersection.groupby([self.user_id])), str(int(remained * 100)) + '%']
            elif self.mode == 'action':
                remained = intersection.shape[0]/first_action_count
                return [intersection.shape[0], str(int(remained * 100)) + '%']
            elif self.mode == 'sum':
                if not self.sum_word:
                    errmsg = 'mode为sum时，请制定sum字段'
                    logging.error(errmsg)
                    raise Exception(print(errmsg))
                else:
                    intersection[self.sum_word] = intersection[self.sum_word].astype('float')
                    intersection_sum = intersection[self.sum_word].sum()
                    return [intersection_sum, str(int(intersection_sum/first_action_sum * 100)) + '%']
        else:
            return [-1, -1]

    # 拿着每组新增用户扫描该组之后按loop的各组，计算每一组的loop留存数组
    def retention(self):
        time_groups = self.data.groupby([self.active_time]).count(
        ).reset_index(self.active_time)[self.active_time].values
        if not self.loop:
            matrix = pd.DataFrame(np.full((len(time_groups), len(
                time_groups)), None), columns=time_groups, index=time_groups)
        else:
            matrix = pd.DataFrame(np.full((len(time_groups), len(
                self.loop)), None), columns=self.loop, index=time_groups)
        matrix_rate = matrix.copy(deep=True)
        for i in range(0, len(time_groups)):
            time = time_groups[i]
            init_users_thistime = self.init_users[self.init_users[self.active_time].str.contains(
                time)]
            follow_up_times = []
            if not self.loop:
                for j in range(i, len(time_groups)):
                    follow_up_times.append(time_groups[j])
            else:
                for j in self.loop:
                    if j + i < len(time_groups):
                        follow_up_times.append(time_groups[i + j])
                    else:
                        break
            first_group = self.data[self.data[self.active_time].str.contains(
                    follow_up_times[0])]
            first_group = pd.merge(init_users_thistime, first_group, on=self.user_id)
            first_action_count = first_group.shape[0]
            if self.mode == 'sum':
                first_action_sum = first_group[self.sum_word].sum()
            else:
                first_action_sum = 0
            for i in range(0, len(follow_up_times)):
                follow_up_time = follow_up_times[i]
                follow_up_users = self.data[self.data[self.active_time].str.contains(
                    follow_up_time)]
                remained = self._remained(df_origin = init_users_thistime, df_current = follow_up_users, first_action_count = first_action_count, first_action_sum = first_action_sum)
                # loc column index
                if not self.loop:
                    _index = follow_up_time
                else:
                    _index = self.loop[i]
                matrix.loc[time, _index] = remained[0]
                matrix_rate.loc[time, _index] = remained[1]
                print(i, follow_up_time)
        matrix.to_csv(self.write_name, encoding='utf_8_sig')
        matrix_rate.to_csv(self.write_name, encoding='utf_8_sig', mode='a')


if __name__ == '__main__':
    # Retention('test.csv', loop=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33], format='%Y/%m/%d %H:%M', period='quarter')
    # Retention('data-all.csv', loop=[0, 1, 2, 3, 4, 5, 6, 7, 8],
    #           format='%Y/%m/%d %H:%M', period='quarter')

    # Retention('data.csv', format='%Y-%m-%d %H:%M:%S',
    #           period='month', write_name='projectg-count-month.csv', mode='count')

    # Retention('data.csv', format='%Y-%m-%d %H:%M:%S',
    #           period='month', write_name='projectg-rate-month.csv', mode='rate')

    Retention(format='%Y-%m-%d %H:%M:%S', loop=[0, 1, 2], period='month', mode='rate', write_name='eg-loop-month-rate.csv')
    Retention(format='%Y-%m-%d %H:%M:%S', period='month', mode='sum', sum_word='order_amount', write_name='eg-noloop-month-sum.csv')
